@@ -1,19 +1,35 @@
 """3D backpropagation algorithm"""
 import ctypes
-import gc
 import multiprocessing as mp
-import platform
 
 import numexpr as ne
 import numpy as np
 import pyfftw
 import scipy.ndimage
 
-import odtbrain
-
 from . import util
 
-_ncores = mp.cpu_count()
+
+ncores = mp.cpu_count()
+mprotate_dict = {}
+
+
+def _cleanup_worker():
+    if "X" in mprotate_dict:
+        mprotate_dict.pop("X")
+    if "X_shape" in mprotate_dict:
+        mprotate_dict.pop("X_shape")
+    if "X_dtype" in mprotate_dict:
+        mprotate_dict.pop("X_dtype")
+
+
+def _init_worker(X, X_shape, X_dtype):
+    """Initializer for pool for _mprotate"""
+    # Using a dictionary is not strictly necessary. You can also
+    # use global variables.
+    mprotate_dict["X"] = X
+    mprotate_dict["X_shape"] = X_shape
+    mprotate_dict["X_dtype"] = X_dtype
 
 
 def _mprotate(ang, lny, pool, order):
@@ -21,8 +37,8 @@ def _mprotate(ang, lny, pool, order):
 
     4x speedup on an intel i7-3820 CPU @ 3.60GHz with 8 cores.
 
-    The function calls _rotate which accesses the
-    `odtbrain._shared_array`. Data is rotated in-place.
+    The function calls _rotate which accesses the `mprotate_dict`.
+    Data is rotated in-place.
 
     Parameters
     ----------
@@ -37,38 +53,29 @@ def _mprotate(ang, lny, pool, order):
     """
     targ_args = list()
 
-    slsize = np.int(np.floor(lny / _ncores))
+    slsize = np.int(np.floor(lny / ncores))
 
-    for t in range(_ncores):
+    for t in range(ncores):
         ymin = t * slsize
         ymax = (t + 1) * slsize
-        if t == _ncores - 1:
+        if t == ncores - 1:
             ymax = lny
         targ_args.append((ymin, ymax, ang, order))
 
-    if platform.system() == "Windows":
-        # Because Windows does not support forking,
-        # the subprocess does not have access to
-        # odtbrain._shared_array. We circumvent
-        # this problem by not using a pool.
-        # We could copy all the data instead and
-        # use a _rotate function that accepts the
-        # array as an argument.
-        for d in targ_args:
-            _rotate(d)
-
-    else:
-        pool.map(_rotate, targ_args)
+    pool.map(_rotate, targ_args)
 
 
 def _rotate(d):
+    arr = np.frombuffer(mprotate_dict["X"],
+                        dtype=mprotate_dict["X_dtype"]).reshape(
+                            mprotate_dict["X_shape"])
     (ymin, ymax, ang, order) = d
     return scipy.ndimage.interpolation.rotate(
-        odtbrain._shared_array[:, ymin:ymax, :],  # input
+        arr[:, ymin:ymax, :],  # input
         angle=-ang,  # angle
         axes=(0, 2),  # axes
         reshape=False,  # reshape
-        output=odtbrain._shared_array[:, ymin:ymax, :],  # output
+        output=arr[:, ymin:ymax, :],  # output
         order=order,  # order
         mode="constant",  # mode
         cval=0)
@@ -76,14 +83,14 @@ def _rotate(d):
 
 def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
                      weight_angles=True, onlyreal=False,
-                     padding=(True, True), padfac=1.75, padval=None,
+                     padding=(True, True), padfac=1.75, padval="edge",
                      intp_order=2, dtype=None,
-                     num_cores=_ncores,
+                     num_cores=ncores,
                      save_memory=False,
                      copy=True,
                      count=None, max_count=None,
                      verbose=0):
-    """3D backpropagation
+    r"""3D backpropagation
 
     Three-dimensional diffraction tomography reconstruction
     algorithm for scattering of a plane wave
@@ -97,22 +104,22 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
 
     .. math::
         f(\mathbf{r}) =
-            -\\frac{i k_\mathrm{m}}{2\pi}
-            \\sum_{j=1}^{N} \! \Delta \phi_0 D_{-\phi_j} \!\!
-            \\left \{
-            \\text{FFT}^{-1}_{\mathrm{2D}}
-            \\left \{
-            \\left| k_\mathrm{Dx} \\right|
-            \\frac{\\text{FFT}_{\mathrm{2D}} \\left \{
-            u_{\mathrm{B},\phi_j}(x_\mathrm{D}, y_\mathrm{D}) \\right \}}
+            -\frac{i k_\mathrm{m}}{2\pi}
+            \sum_{j=1}^{N} \! \Delta \phi_0 D_{-\phi_j} \!\!
+            \left \{
+            \text{FFT}^{-1}_{\mathrm{2D}}
+            \left \{
+            \left| k_\mathrm{Dx} \right|
+            \frac{\text{FFT}_{\mathrm{2D}} \left \{
+            u_{\mathrm{B},\phi_j}(x_\mathrm{D}, y_\mathrm{D}) \right \}}
             {u_0(l_\mathrm{D})}
-            \exp \! \\left[i k_\mathrm{m}(M - 1) \cdot
-            (z_{\phi_j}-l_\mathrm{D}) \\right]
-            \\right \}
-            \\right \}
+            \exp \! \left[i k_\mathrm{m}(M - 1) \cdot
+            (z_{\phi_j}-l_\mathrm{D}) \right]
+            \right \}
+            \right \}
 
-    with the forward :math:`\\text{FFT}_{\mathrm{2D}}` and inverse
-    :math:`\\text{FFT}^{-1}_{\mathrm{2D}}` 2D fast Fourier transform, the
+    with the forward :math:`\text{FFT}_{\mathrm{2D}}` and inverse
+    :math:`\text{FFT}^{-1}_{\mathrm{2D}}` 2D fast Fourier transform, the
     rotational operator :math:`D_{-\phi_j}`, the angular distance between the
     projections :math:`\Delta \phi_0`, the ramp filter in Fourier space
     :math:`|k_\mathrm{Dx}|`, and the propagation distance
@@ -144,8 +151,8 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
         projections.
 
         .. math::
-            \Delta \phi_0 \\longmapsto \Delta \phi_j =
-                \\frac{\phi_{j+1} - \phi_{j-1}}{2}
+            \Delta \phi_0 \longmapsto \Delta \phi_j =
+                \frac{\phi_{j+1} - \phi_{j-1}}{2}
 
         .. versionadded:: 0.1.1
     onlyreal: bool
@@ -167,13 +174,14 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
         lead to a padded size of 512 for an initial size of 150.
         Values geater than 2 are allowed. This parameter may
         greatly increase memory usage!
-    padval: float
+    padval: float or "edge"
         The value used for padding. This is important for the Rytov
         approximation, where an approximat zero in the phase might
         translate to 2πi due to the unwrapping algorithm. In that
         case, this value should be a multiple of 2πi.
-        If `padval` is `None`, then the edge values are used for
-        padding (see documentation of :func:`numpy.pad`).
+        If `padval` is "edge", then the edge values are used for
+        padding (see documentation of :func:`numpy.pad`). If `padval`
+        is a float, then padding is done with a linear ramp.
     intp_order: int between 0 and 5
         Order of the interpolation for rotation.
         See :func:`scipy.ndimage.interpolation.rotate` for details.
@@ -209,7 +217,7 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
         Reconstructed object function :math:`f(\mathbf{r})` as defined
         by the Helmholtz equation.
         :math:`f(x,z) =
-        k_m^2 \\left(\\left(\\frac{n(x,z)}{n_m}\\right)^2 -1\\right)`
+        k_m^2 \left(\left(\frac{n(x,z)}{n_m}\right)^2 -1\right)`
 
 
     See Also
@@ -226,43 +234,83 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
     with a numerical focusing algorithm (available in the Python
     package :py:mod:`nrefocus`).
     """
-    ne.set_num_threads(num_cores)
+    A = angles.size
 
-    if copy:
-        uSin = uSin.copy()
+    if len(uSin.shape) != 3:
+        raise ValueError("Input data `uSin` must have shape (A,Ny,Nx).")
+    if len(uSin) != A:
+        raise ValueError("`len(angles)` must be  equal to `len(uSin)`.")
+    if len(list(padding)) != 2:
+        raise ValueError("`padding` must be boolean tuple of length 2!")
+    if np.array(padding).dtype is not np.dtype(bool):
+        raise ValueError("Parameter `padding` must be boolean tuple.")
+    if coords is not None:
+        raise NotImplementedError("Setting coordinates is not yet supported.")
+    if num_cores > ncores:
+        raise ValueError("`num_cores` must not exceed number "
+                         + "of physical cores: {}".format(ncores))
 
-    A = angles.shape[0]
-    # jobmanager
-    if max_count is not None:
-        max_count.value += A + 2
-
-    # check for dtype
+    # setup dtype
     if dtype is None:
         dtype = np.float_
     dtype = np.dtype(dtype)
-    assert dtype.name in ["float32",
-                          "float64"], "dtype must be float32 or float64!"
-
-    assert num_cores <= _ncores, "`num_cores` must not exceed number " +\
-                                 "of physical cores: {}".format(_ncores)
-
-    assert np.iscomplexobj(uSin), "uSin dtype must be complex128."
-
+    if dtype.name not in ["float32", "float64"]:
+        raise ValueError("dtype must be float32 or float64!")
     dtype_complex = np.dtype("complex{}".format(
         2 * np.int(dtype.name.strip("float"))))
-
     # set ctype
     ct_dt_map = {np.dtype(np.float32): ctypes.c_float,
                  np.dtype(np.float64): ctypes.c_double
                  }
 
-    assert len(uSin.shape) == 3, "Input data `uSin` must have shape (A,Ny,Nx)."
-    assert len(uSin) == A, "`len(angles)` must be  equal to `len(uSin)`."
-    assert len(
-        list(padding)) == 2, "`padding` must be boolean tuple of length 2!"
-    assert np.array(padding).dtype is np.dtype(
-        bool), "Parameter `padding` must be boolean tuple."
-    assert coords is None, "Setting coordinates is not yet supported."
+    # progress
+    if max_count is not None:
+        max_count.value += A + 2
+
+    ne.set_num_threads(num_cores)
+
+    uSin = np.array(uSin, copy=copy)
+
+    # lengths of the input data
+    lny, lnx = uSin.shape[1], uSin.shape[2]
+    # The z-size of the output array must match the x-size.
+    # The rotation is performed about the y-axis (lny).
+    ln = lnx
+
+    # We perform zero-padding before performing the Fourier transform.
+    # This gets rid of artifacts due to false periodicity and also
+    # speeds up Fourier transforms of the input image size is not
+    # a power of 2.
+
+    if padding[0]:
+        orderx = np.int(max(64., 2**np.ceil(np.log(lnx * padfac) / np.log(2))))
+        padx = orderx - lnx
+    else:
+        padx = 0
+    if padding[1]:
+        ordery = np.int(max(64., 2**np.ceil(np.log(lny * padfac) / np.log(2))))
+        pady = ordery - lny
+    else:
+        pady = 0
+
+    padyl = np.int(np.ceil(pady / 2))
+    padyr = pady - padyl
+    padxl = np.int(np.ceil(padx / 2))
+    padxr = padx - padxl
+
+    # zero-padded length of sinogram.
+    lNx = lnx + padx
+    lNy = lny + pady
+    lNz = ln
+
+    if verbose > 0:
+        print("......Image size (x,y): {}x{}, padded: {}x{}".format(
+            lnx, lny, lNx, lNy))
+
+    # Perform weighting
+    if weight_angles:
+        weights = util.compute_angle_weights_1d(angles).reshape(-1, 1, 1)
+        uSin *= weights
 
     # Cut-Off frequency
     # km [1/px]
@@ -282,70 +330,6 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
     # This is not a big problem. We only need to multiply the imaginary
     # part of the scattered wave by -1.
 
-    sinogram = uSin
-    # Perform weighting
-    if weight_angles:
-        weights = util.compute_angle_weights_1d(angles).reshape(-1, 1, 1)
-        sinogram *= weights
-
-    # lengths of the input data
-    (la, lny, lnx) = sinogram.shape
-    # The z-size of the output array must match the x-size.
-    # The rotation is performed about the y-axis (lny).
-    ln = lnx
-
-    # We perform padding before performing the Fourier transform.
-    # This gets rid of artifacts due to false periodicity and also
-    # speeds up Fourier transforms of the input image size is not
-    # a power of 2.
-    # transpose so we can call resize correctly
-
-    orderx = max(64., 2**np.ceil(np.log(lnx * padfac) / np.log(2)))
-    ordery = max(64., 2**np.ceil(np.log(lny * padfac) / np.log(2)))
-
-    if padding[0]:
-        padx = orderx - lnx
-    else:
-        padx = 0
-    if padding[1]:
-        pady = ordery - lny
-    else:
-        pady = 0
-
-    # Apply a Fourier filter before projecting the sinogram slices.
-    # Resize image to next power of two for fourier analysis
-    # Reduces artifacts
-
-    padyl = np.int(np.ceil(pady / 2))
-    padyr = np.int(pady - padyl)
-    padxl = np.int(np.ceil(padx / 2))
-    padxr = np.int(padx - padxl)
-
-    # TODO: This padding takes up a lot of memory. Move it to a separate
-    # for loop or to the main for-loop.
-    if padval is None:
-        sino = np.pad(sinogram, ((0, 0), (padyl, padyr), (padxl, padxr)),
-                      mode="edge")
-        if verbose > 0:
-            print("......Padding with edge values.")
-    else:
-        sino = np.pad(sinogram, ((0, 0), (padyl, padyr), (padxl, padxr)),
-                      mode="linear_ramp",
-                      end_values=(padval,))
-        if verbose > 0:
-            print("......Verifying padding value: {}".format(padval))
-
-    # save memory
-    del sinogram
-
-    if verbose > 0:
-        print("......Image size (x,y): {}x{}, padded: {}x{}".format(
-            lnx, lny, sino.shape[2], sino.shape[1]))
-
-    # zero-padded length of sinogram.
-    (lA, lNy, lNx) = sino.shape  # @UnusedVariable
-    lNz = ln
-
     # Ask for the filter. Do not include zero (first element).
     #
     # Integrals over ϕ₀ [0,2π]; kx [-kₘ,kₘ]
@@ -392,9 +376,9 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
     # Differentials for integral
     dphi0 = 2 * np.pi / A
     # We will later multiply with phi0.
-    #               a, y, x
-    kx = kx.reshape(1, 1, -1)
-    ky = ky.reshape(1, -1, 1)
+    #               y, x
+    kx = kx.reshape(1, -1)
+    ky = ky.reshape(-1, 1)
     # Low-pass filter:
     # less-than-or-equal would give us zero division error.
     filter_klp = (kx**2 + ky**2 < km**2)
@@ -413,40 +397,10 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
     # to take into account that we have a scattered
     # wave that is normalized by u0.
     prefactor *= np.exp(-1j * km * (M-1) * lD)
-    # Perform filtering of the sinogram,
-    # save memory by in-place operations
-    # projection = np.fft.fft2(sino, axes=(-1,-2)) * prefactor
-    # FFTW-flag is "estimate":
-    #   specifies that, instead of actual measurements of different
-    #   algorithms, a simple heuristic is used to pick a (probably
-    #   sub-optimal) plan quickly. With this flag, the input/output
-    #   arrays are not overwritten during planning.
-
-    # Byte-aligned arrays
-    temp_array = pyfftw.n_byte_align_empty(sino[0].shape, 16, dtype_complex)
-
-    myfftw_plan = pyfftw.FFTW(temp_array, temp_array, threads=num_cores,
-                              flags=["FFTW_ESTIMATE"], axes=(0, 1))
 
     if count is not None:
         count.value += 1
 
-    for p in range(len(sino)):
-        # this overwrites sino
-        temp_array[:] = sino[p, :, :]
-        myfftw_plan.execute()
-        sino[p, :, :] = temp_array[:]
-
-    temp_array, myfftw_plan
-
-    projection = sino
-    # - normalize to (lNx * lNy) for FFTW
-    projection[:] *= prefactor / (lNx * lNy)
-
-    # save memory
-    del prefactor, filter_klp
-    #
-    #
     # filter (2) must be applied before rotation as well
     # exp( i (kx t⊥ + kₘ (M - 1) s₀) r )
     #
@@ -479,29 +433,37 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
         # compute filter2 now
         filter2 = ne.evaluate("exp(factor * zv)",
                               local_dict={"factor": f2_exp_fac,
-                                          "zv": zv})
+                                          "zv": zv},
+                              casting="same_kind")
         # occupies some amount of ram, but yields faster
         # computation later
-        # filter2[0].size*len(filter2)*128/8/1024**3
 
     if count is not None:
         count.value += 1
 
-    #                                  a, z, y,  x
-    # projection = projection.reshape(la, 1, lNy, lNx)
-    projection = projection.reshape(la, lNy, lNx)
-
-    # This frees comparatively few data
-    del M
-
     # Prepare complex output image
     if onlyreal:
         outarr = np.zeros((ln, lny, lnx), dtype=dtype)
     else:
         outarr = np.zeros((ln, lny, lnx), dtype=dtype_complex)
 
-    # Create plan for fftw:
-    inarr = pyfftw.n_byte_align_empty((lNy, lNx), 16, dtype_complex)
+    # Create plan for FFTW
+    # save memory by in-place operations
+    # projection = np.fft.fft2(sino, axes=(-1,-2)) * prefactor
+    # FFTW-flag is "estimate":
+    #   specifies that, instead of actual measurements of different
+    #   algorithms, a simple heuristic is used to pick a (probably
+    #   sub-optimal) plan quickly. With this flag, the input/output
+    #   arrays are not overwritten during planning.
+
+    # Byte-aligned arrays
+    oneslice = pyfftw.empty_aligned((lNy, lNx), dtype_complex)
+
+    myfftw_plan = pyfftw.FFTW(oneslice, oneslice, threads=num_cores,
+                              flags=["FFTW_ESTIMATE"], axes=(0, 1))
+
+    # Create plan for IFFTW:
+    inarr = pyfftw.empty_aligned((lNy, lNx), dtype_complex)
     # inarr[:] = (projection[0]*filter2)[0,:,:]
     # plan is "patient":
     #    FFTW_PATIENT is like FFTW_MEASURE, but considers a wider range
@@ -516,54 +478,59 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
                                direction="FFTW_BACKWARD",
                                flags=["FFTW_MEASURE"])
 
-    # assert shared_array.base.base is shared_array_base.get_obj()
-    shared_array_base = mp.Array(ct_dt_map[dtype], ln * lny * lnx)
-    _shared_array = np.ctypeslib.as_array(shared_array_base.get_obj())
-    _shared_array = _shared_array.reshape(ln, lny, lnx)
+    # Setup a shared array
+    shared_array = mp.RawArray(ct_dt_map[dtype], ln * lny * lnx)
+    arr = np.frombuffer(shared_array, dtype=dtype).reshape(ln, lny, lnx)
 
     # Initialize the pool with the shared array
-    odtbrain._shared_array = _shared_array
-    pool4loop = mp.Pool(processes=num_cores)
+    pool4loop = mp.Pool(processes=num_cores,
+                        initializer=_init_worker,
+                        initargs=(shared_array, (ln, lny, lnx), dtype))
 
     # filtered projections in loop
     filtered_proj = np.zeros((ln, lny, lnx), dtype=dtype_complex)
 
     for aa in np.arange(A):
+        if not (padding[0] and padding[1]):
+            # no padding
+            oneslice[:] = uSin[aa]
+        elif padval == "edge":
+            # padding with edge values
+            oneslice[:] = np.pad(uSin[aa],
+                                 ((padyl, padyr), (padxl, padxr)),
+                                 mode="edge")
+        else:
+            # padding with linear ramp
+            oneslice[:] = np.pad(uSin[aa],
+                                 ((padyl, padyr), (padxl, padxr)),
+                                 mode="linear_ramp",
+                                 end_values=(padval,))
+        myfftw_plan.execute()
+        # normalize to (lNx * lNy) for FFTW and multiply with prefactor
+        oneslice *= prefactor / (lNx * lNy)
         # 14x Speedup with fftw3 compared to numpy fft and
         # memory reduction by a factor of 2!
         # ifft will be computed in-place
 
-        # A == la
-        # projection.shape == (A, lNx, lNy)
-        # filter2.shape == (ln, lNx, lNy)
         for p in range(len(zv)):
             if save_memory:
                 # compute filter2 here;
                 # this is comparatively slower than the other case
                 ne.evaluate("exp(factor * zvp) * projectioni",
                             local_dict={"zvp": zv[p],
-                                        "projectioni": projection[aa],
+                                        "projectioni": oneslice,
                                         "factor": f2_exp_fac},
+                            casting="same_kind",
                             out=inarr)
             else:
                 # use universal functions
-                np.multiply(filter2[p], projection[aa], out=inarr)
+                np.multiply(filter2[p], oneslice, out=inarr)
             myifftw_plan.execute()
-            filtered_proj[p, :, :] = inarr[
-                padyl:padyl + lny,
-                padxl:padxl + lnx
-            ]
+            filtered_proj[p, :, :] = inarr[padyl:lny+padyl, padxl:lnx+padxl]
 
         # resize image to original size
         # The copy is necessary to prevent memory leakage.
-        # The fftw did not normalize the data.
-        # _shared_array[:] = sino_filtered.real[:ln, :lny, :lnx] / (lNx * lNy)
-        # By performing the "/" operation here, we magically use less
-        # memory and we gain speed...
-        _shared_array[:] = filtered_proj.real
-        # _shared_array[:] = sino_filtered.real[:ln,
-        #                                       padyl:padyl+lny,
-        #                                       padxl:padxl+lnx] / (lNx * lNy)
+        arr[:] = filtered_proj.real
 
         phi0 = np.rad2deg(angles[aa])
 
@@ -572,25 +539,18 @@ def backpropagate_3d(uSin, angles, res, nm, lD=0, coords=None,
 
         _mprotate(phi0, lny, pool4loop, intp_order)
 
-        outarr.real += _shared_array
+        outarr.real += arr
 
         if not onlyreal:
-            _shared_array[:] = filtered_proj_imag
-            # _shared_array[:] = sino_filtered_imag[
-            #    :ln, :lny, :lnx] / (lNx * lNy)
-            del filtered_proj_imag
+            arr[:] = filtered_proj_imag
             _mprotate(phi0, lny, pool4loop, intp_order)
-            outarr.imag += _shared_array
+            outarr.imag += arr
 
         if count is not None:
             count.value += 1
 
     pool4loop.terminate()
     pool4loop.join()
-
-    del _shared_array, inarr, odtbrain._shared_array
-    del shared_array_base
-
-    gc.collect()
+    _cleanup_worker()
 
     return outarr
