@@ -1,6 +1,4 @@
 """3D backpropagation algorithm with a tilted axis of rotation"""
-import ctypes
-import gc
 import multiprocessing as mp
 import warnings
 
@@ -9,10 +7,10 @@ import numpy as np
 import pyfftw
 import scipy.ndimage
 
-
-from ._alg3d_bpp import _ncores
 from . import util
-import odtbrain
+
+
+_ncores = mp.cpu_count()
 
 
 def estimate_major_rotation_axis(loc):
@@ -293,7 +291,12 @@ def sphere_points_from_angles_and_tilt(angles, tilted_axis):
 
     # (b) This is the polar angle measured in the x-z plane starting
     #     at the x-axis and measured towards the positive z-axis.
-    phi = np.arctan2(u, w)
+    if np.allclose(u, 0) and np.allclose(w, 0):
+        # Avoid flipping the axis of rotation due to numerical
+        # errors during its computation.
+        phi = 0
+    else:
+        phi = np.arctan2(u, w)
 
     # Determine the projection points on the unit sphere.
     # The resulting circle meets the x-z-plane at phi, and
@@ -371,14 +374,14 @@ def sphere_points_from_angles_and_tilt(angles, tilted_axis):
 def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
                             tilted_axis=[0, 1, 0],
                             coords=None, weight_angles=True, onlyreal=False,
-                            padding=(True, True), padfac=1.75, padval=None,
+                            padding=(True, True), padfac=1.75, padval="edge",
                             intp_order=2, dtype=None,
                             num_cores=_ncores,
                             save_memory=False,
                             copy=True,
                             count=None, max_count=None,
                             verbose=0):
-    """3D backpropagation with a tilted axis of rotation
+    r"""3D backpropagation with a tilted axis of rotation
 
     Three-dimensional diffraction tomography reconstruction
     algorithm for scattering of a plane wave
@@ -387,28 +390,28 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     :math:`n(x,y,z)`.
 
     This method implements the 3D backpropagation algorithm with
-    a rotational axis that is tilted by :math:`\\theta_\mathrm{tilt}`
+    a rotational axis that is tilted by :math:`\theta_\mathrm{tilt}`
     w.r.t. the imaging plane :cite:`Mueller2015tilted`.
 
     .. math::
         f(\mathbf{r}) =
-            -\\frac{i k_\mathrm{m}}{2\pi}
-            \\sum_{j=1}^{N} \! \Delta \phi_0 D_{-\phi_j}^\mathrm{tilt} \!\!
-            \\left \{
-            \\text{FFT}^{-1}_{\mathrm{2D}}
-            \\left \{
-            \\left| k_\mathrm{Dx} \cdot \cos \\theta_\mathrm{tilt}\\right|
-            \\frac{\\text{FFT}_{\mathrm{2D}} \\left \{
-            u_{\mathrm{B},\phi_j}(x_\mathrm{D}, y_\mathrm{D}) \\right \}}
+            -\frac{i k_\mathrm{m}}{2\pi}
+            \sum_{j=1}^{N} \! \Delta \phi_0 D_{-\phi_j}^\mathrm{tilt} \!\!
+            \left \{
+            \text{FFT}^{-1}_{\mathrm{2D}}
+            \left \{
+            \left| k_\mathrm{Dx} \cdot \cos \theta_\mathrm{tilt}\right|
+            \frac{\text{FFT}_{\mathrm{2D}} \left \{
+            u_{\mathrm{B},\phi_j}(x_\mathrm{D}, y_\mathrm{D}) \right \}}
             {u_0(l_\mathrm{D})}
-            \exp \! \\left[i k_\mathrm{m}(M - 1) \cdot
-            (z_{\phi_j}-l_\mathrm{D}) \\right]
-            \\right \}
-            \\right \}
+            \exp \! \left[i k_\mathrm{m}(M - 1) \cdot
+            (z_{\phi_j}-l_\mathrm{D}) \right]
+            \right \}
+            \right \}
 
     with a modified rotational operator :math:`D_{-\phi_j}^\mathrm{tilt}`
     and a different filter in Fourier space
-    :math:`|k_\mathrm{Dx} \cdot \cos \\theta_\mathrm{tilt}|` when compared
+    :math:`|k_\mathrm{Dx} \cdot \cos \theta_\mathrm{tilt}|` when compared
     to :func:`backpropagate_3d`.
 
 
@@ -451,8 +454,8 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
         projections.
 
         .. math::
-            \Delta \phi_0 \\longmapsto \Delta \phi_j =
-                \\frac{\phi_{j+1} - \phi_{j-1}}{2}
+            \Delta \phi_0 \longmapsto \Delta \phi_j =
+                \frac{\phi_{j+1} - \phi_{j-1}}{2}
 
         This currently only works when `angles` has the shape (A,).
     onlyreal: bool
@@ -470,17 +473,18 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
         Increase padding size of the input data. A value greater
         than one will trigger padding to the second-next power of
         two. For example, a value of 1.75 will lead to a padded
-        size of 256 for an initial size of 144, whereas for it will
+        size of 256 for an initial size of 144, whereas it will
         lead to a padded size of 512 for an initial size of 150.
-        Values geater than 2 are allowed. This parameter may
+        Values greater than 2 are allowed. This parameter may
         greatly increase memory usage!
-    padval: float
+    padval: float or "edge"
         The value used for padding. This is important for the Rytov
-        approximation, where an approximat zero in the phase might
+        approximation, where an approximate zero in the phase might
         translate to 2πi due to the unwrapping algorithm. In that
         case, this value should be a multiple of 2πi.
-        If `padval` is `None`, then the edge values are used for
-        padding (see documentation of :func:`numpy.pad`).
+        If `padval` is "edge", then the edge values are used for
+        padding (see documentation of :func:`numpy.pad`). If `padval`
+        is a float, then padding is done with a linear ramp.
     intp_order: int between 0 and 5
         Order of the interpolation for rotation.
         See :func:`scipy.ndimage.interpolation.affine_transform` for details.
@@ -516,7 +520,7 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
         Reconstructed object function :math:`f(\mathbf{r})` as defined
         by the Helmholtz equation.
         :math:`f(x,z) =
-        k_m^2 \\left(\\left(\\frac{n(x,z)}{n_m}\\right)^2 -1\\right)`
+        k_m^2 \left(\left(\frac{n(x,z)}{n_m}\right)^2 -1\right)`
 
 
     See Also
@@ -555,11 +559,75 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     with a numerical focusing algorithm (available in the Python
     package :py:mod:`nrefocus`).
     """
+    A = angles.shape[0]
+
+    if angles.shape not in [(A,), (A, 1), (A, 3)]:
+        raise ValueError("`angles` must have shape (A,) or (A,3)!")
+    if len(uSin.shape) != 3:
+        raise ValueError("Input data `uSin` must have shape (A,Ny,Nx).")
+    if len(uSin) != A:
+        raise ValueError("`len(angles)` must be  equal to `len(uSin)`.")
+    if len(list(padding)) != 2:
+        raise ValueError("`padding` must be boolean tuple of length 2!")
+    if np.array(padding).dtype is not np.dtype(bool):
+        raise ValueError("Parameter `padding` must be boolean tuple.")
+    if coords is not None:
+        raise NotImplementedError("Setting coordinates is not yet supported.")
+    if num_cores > _ncores:
+        raise ValueError("`num_cores` must not exceed number "
+                         + "of physical cores: {}".format(_ncores))
+
+    # setup dtype
+    if dtype is None:
+        dtype = np.float_
+    dtype = np.dtype(dtype)
+    if dtype.name not in ["float32", "float64"]:
+        raise ValueError("dtype must be float32 or float64!")
+    dtype_complex = np.dtype("complex{}".format(
+        2 * int(dtype.name.strip("float"))))
+
+    # progess monitoring
+    if max_count is not None:
+        max_count.value += A + 2
+
     ne.set_num_threads(num_cores)
 
-    if copy:
-        uSin = uSin.copy()
-        angles = angles.copy()
+    uSin = np.array(uSin, copy=copy)
+    angles = np.array(angles, copy=copy)
+    angles = np.squeeze(angles)  # support shape (A,1)
+
+    # lengths of the input data
+    lny, lnx = uSin.shape[1], uSin.shape[2]
+    ln = lnx
+
+    # We perform zero-padding before performing the Fourier transform.
+    # This gets rid of artifacts due to false periodicity and also
+    # speeds up Fourier transforms of the input image size is not
+    # a power of 2.
+    if padding[0]:
+        orderx = np.int(max(64., 2**np.ceil(np.log(lnx * padfac) / np.log(2))))
+        padx = orderx - lnx
+    else:
+        padx = 0
+    if padding[1]:
+        ordery = np.int(max(64., 2**np.ceil(np.log(lny * padfac) / np.log(2))))
+        pady = ordery - lny
+    else:
+        pady = 0
+
+    padyl = np.int(np.ceil(pady / 2))
+    padyr = pady - padyl
+    padxl = np.int(np.ceil(padx / 2))
+    padxr = padx - padxl
+
+    # zero-padded length of sinogram.
+    lNx = lnx + padx
+    lNy = lny + pady
+    lNz = ln
+
+    if verbose > 0:
+        print("......Image size (x,y): {}x{}, padded: {}x{}".format(
+            lnx, lny, lNx, lNy))
 
     # `tilted_axis` is required for several things:
     # 1. the filter |kDx*v + kDy*u| with (u,v,w)==tilted_axis
@@ -588,14 +656,6 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     # rotate `tilted_axis` onto the y-z plane.
     tilted_axis_yz = norm_vec(np.dot(rotmat, tilted_axis))
 
-    A = angles.shape[0]
-    angles = np.squeeze(angles)  # Allow shapes (A,1)
-    assert angles.shape in [
-        (A,), (A, 3)], "`angles` must have shape (A,) or (A,3)!"
-    # jobmanager
-    if max_count is not None:
-        max_count.value += A + 2
-
     if len(angles.shape) == 1:
         if weight_angles:
             weights = util.compute_angle_weights_1d(angles).reshape(-1, 1, 1)
@@ -612,33 +672,8 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
             # instead rotate like `tilted_axis` onto the y-z plane.
             angles[ii] = norm_vec(np.dot(rotmat, angles[ii]))
 
-    # check for dtype
-    if dtype is None:
-        dtype = np.float_
-    dtype = np.dtype(dtype)
-    assert dtype.name in ["float32",
-                          "float64"], "dtype must be float32 or float64!"
-
-    assert num_cores <= _ncores, "`num_cores` must not exceed number " +\
-                                 "of physical cores: {}".format(_ncores)
-
-    assert np.iscomplexobj(uSin), "uSin dtype must be complex128."
-
-    dtype_complex = np.dtype("complex{}".format(
-        2 * int(dtype.name.strip("float"))))
-
-    # set ctype
-    ct_dt_map = {np.dtype(np.float32): ctypes.c_float,
-                 np.dtype(np.float64): ctypes.c_double
-                 }
-
-    assert len(uSin.shape) == 3, "Input data `uSin` must have shape (A,Ny,Nx)."
-    assert len(uSin) == A, "`len(angles)` must be  equal to `len(uSin)`."
-    assert len(
-        list(padding)) == 2, "`padding` must be boolean tuple of length 2!"
-    assert np.array(padding).dtype is np.dtype(
-        bool), "Parameter `padding` must be boolean tuple."
-    assert coords is None, "Setting coordinates is not yet supported."
+    if weight_angles:
+        uSin *= weights
 
     # Cut-Off frequency
     # km [1/px]
@@ -657,65 +692,6 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     # latter sign convention.
     # This is not a big problem. We only need to multiply the imaginary
     # part of the scattered wave by -1.
-    sinogram = uSin
-
-    if weight_angles:
-        sinogram *= weights
-
-    # lengths of the input data
-    (la, lny, lnx) = sinogram.shape
-    ln = lnx
-
-    # We do a zero-padding before performing the Fourier transform.
-    # This gets rid of artifacts due to false periodicity and also
-    # speeds up Fourier transforms of the input image size is not
-    # a power of 2.
-    # transpose so we can call resize correctly
-
-    orderx = max(64., 2**np.ceil(np.log(lnx * padfac) / np.log(2)))
-    ordery = max(64., 2**np.ceil(np.log(lny * padfac) / np.log(2)))
-
-    if padding[0]:
-        padx = orderx - lnx
-    else:
-        padx = 0
-    if padding[1]:
-        pady = ordery - lny
-    else:
-        pady = 0
-
-    # Apply a Fourier filter before projecting the sinogram slices.
-    # Resize image to next power of two for fourier analysis
-    # (reduces artifacts).
-    padyl = np.int(np.ceil(pady / 2))
-    padyr = np.int(pady - padyl)
-    padxl = np.int(np.ceil(padx / 2))
-    padxr = np.int(padx - padxl)
-
-    # TODO: This padding takes up a lot of memory. Move it to a separate
-    # for loop or to the main for-loop.
-    if padval is None:
-        sino = np.pad(sinogram, ((0, 0), (padyl, padyr), (padxl, padxr)),
-                      mode="edge")
-        if verbose > 0:
-            print("......Padding with edge values.")
-    else:
-        sino = np.pad(sinogram, ((0, 0), (padyl, padyr), (padxl, padxr)),
-                      mode="linear_ramp",
-                      end_values=(padval,))
-        if verbose > 0:
-            print("......Verifying padding value: {}".format(padval))
-
-    # save memory
-    del sinogram
-
-    if verbose > 0:
-        print("......Image size (x,y): {}x{}, padded: {}x{}".format(
-            lnx, lny, sino.shape[2], sino.shape[1]))
-
-    # zero-padded length of sinogram.
-    (lA, lNy, lNx) = sino.shape  # @UnusedVariable
-    lNz = ln
 
     # Ask for the filter. Do not include zero (first element).
     #
@@ -767,8 +743,8 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     dphi0 = 2 * np.pi / A
     # We will later multiply with phi0.
     #               a, y, x
-    kx = kx.reshape(1, 1, -1)
-    ky = ky.reshape(1, -1, 1)
+    kx = kx.reshape(1, -1)
+    ky = ky.reshape(-1, 1)
     # Low-pass filter:
     # less-than-or-equal would give us zero division error.
     filter_klp = (kx**2 + ky**2 < km**2)
@@ -793,38 +769,10 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     # to take into account that we have a scattered
     # wave that is normalized by u0.
     prefactor *= np.exp(-1j * km * (M-1) * lD)
-    # Perform filtering of the sinogram,
-    # save memory by in-place operations
-    # projection = np.fft.fft2(sino, axes=(-1,-2)) * prefactor
-    # Flag is "estimate":
-    #   specifies that, instead of actual measurements of different
-    #   algorithms, a simple heuristic is used to pick a (probably
-    #   sub-optimal) plan quickly. With this flag, the input/output
-    #   arrays are not overwritten during planning.
-
-    # Byte-aligned arrays
-    temp_array = pyfftw.n_byte_align_empty(sino[0].shape, 16, dtype_complex)
-
-    myfftw_plan = pyfftw.FFTW(temp_array, temp_array, threads=num_cores,
-                              flags=["FFTW_ESTIMATE"], axes=(0, 1))
 
     if count is not None:
         count.value += 1
 
-    for p in range(len(sino)):
-        # this overwrites sino
-        temp_array[:] = sino[p, :, :]
-        myfftw_plan.execute()
-        sino[p, :, :] = temp_array[:]
-
-    temp_array, myfftw_plan
-
-    projection = sino
-    projection[:] *= prefactor / (lNx * lNy)
-    projection[:] *= filterabs
-
-    # save memory
-    del prefactor, filter_klp
     #
     #
     # filter (2) must be applied before rotation as well
@@ -855,7 +803,7 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     z = np.linspace(-center, center, lNz, endpoint=False)
     zv = z.reshape(-1, 1, 1)
 
-    #              z, y, x
+    #              y,   x
     Mp = M.reshape(lNy, lNx)
 
     # filter2 = np.exp(1j * zv * km * (Mp - 1))
@@ -865,32 +813,36 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
         pass
     else:
         # compute filter2 now
+        # (this requires more RAM but is faster)
         filter2 = ne.evaluate("exp(factor * zv)",
                               local_dict={"factor": f2_exp_fac,
-                                          "zv": zv})
-        # occupies some amount of ram, but yields faster
-        # computation later
-        # filter2[0].size*len(filter2)*128/8/1024**3
+                                          "zv": zv},
+                              casting="same_kind")
 
     if count is not None:
         count.value += 1
 
-    #                                  a, z, y,  x
-    # projection = projection.reshape(la, 1, lNy, lNx)
-    projection = projection.reshape(la, lNy, lNx)
-
-    # This frees comparatively few data
-    del M
-
     # Prepare complex output image
     if onlyreal:
         outarr = np.zeros((ln, lny, lnx), dtype=dtype)
     else:
         outarr = np.zeros((ln, lny, lnx), dtype=dtype_complex)
 
-    # Create plan for fftw:
-    inarr = pyfftw.n_byte_align_empty((lNy, lNx), 16, dtype_complex)
-    # inarr[:] = (projection[0]*filter2)[0,:,:]
+    # Create plan for FFTW:
+    # Flag is "estimate":
+    #   specifies that, instead of actual measurements of different
+    #   algorithms, a simple heuristic is used to pick a (probably
+    #   sub-optimal) plan quickly. With this flag, the input/output
+    #   arrays are not overwritten during planning.
+
+    # Byte-aligned arrays
+    oneslice = pyfftw.empty_aligned((lNy, lNx), dtype_complex)
+
+    myfftw_plan = pyfftw.FFTW(oneslice, oneslice, threads=num_cores,
+                              flags=["FFTW_ESTIMATE"], axes=(0, 1))
+
+    # Create plan for IFFTW:
+    inarr = pyfftw.empty_aligned((lNy, lNx), dtype_complex)
     # plan is "patient":
     #    FFTW_PATIENT is like FFTW_MEASURE, but considers a wider range
     #    of algorithms and often produces a “more optimal” plan
@@ -904,15 +856,6 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
                                direction="FFTW_BACKWARD",
                                flags=["FFTW_MEASURE"])
 
-    # assert shared_array.base.base is shared_array_base.get_obj()
-    shared_array_base = mp.Array(ct_dt_map[dtype], ln * lny * lnx)
-    _shared_array = np.ctypeslib.as_array(shared_array_base.get_obj())
-    _shared_array = _shared_array.reshape(ln, lny, lnx)
-
-    # Initialize the pool with the shared array
-    odtbrain._shared_array = _shared_array
-    pool4loop = mp.Pool(processes=num_cores)
-
     # filtered projections in loop
     filtered_proj = np.zeros((ln, lny, lnx), dtype=dtype_complex)
 
@@ -921,9 +864,23 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
     angles = rotate_points_to_axis(points=angles, axis=tilted_axis_yz)
 
     for aa in np.arange(A):
-        # A == la
-        # projection.shape == (A, lNx, lNy)
-        # filter2.shape == (ln, lNx, lNy)
+        if not (padding[0] and padding[1]):
+            # no padding
+            oneslice[:] = uSin[aa]
+        elif padval == "edge":
+            # padding with edge values
+            oneslice[:] = np.pad(uSin[aa],
+                                 ((padyl, padyr), (padxl, padxr)),
+                                 mode="edge")
+        else:
+            # padding with linear ramp
+            oneslice[:] = np.pad(uSin[aa],
+                                 ((padyl, padyr), (padxl, padxr)),
+                                 mode="linear_ramp",
+                                 end_values=(padval,))
+        myfftw_plan.execute()
+        # normalize to (lNx * lNy) for FFTW and multiply with prefactor, filter
+        oneslice *= filterabs * prefactor / (lNx * lNy)
 
         for p in range(len(zv)):
             if save_memory:
@@ -931,17 +888,15 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
                 # this is comparatively slower than the other case
                 ne.evaluate("exp(factor * zvp) * projectioni",
                             local_dict={"zvp": zv[p],
-                                        "projectioni": projection[aa],
+                                        "projectioni": oneslice,
                                         "factor": f2_exp_fac},
+                            casting="same_kind",
                             out=inarr)
             else:
                 # use universal functions
-                np.multiply(filter2[p], projection[aa], out=inarr)
+                np.multiply(filter2[p], oneslice, out=inarr)
             myifftw_plan.execute()
-            filtered_proj[p, :, :] = inarr[
-                padyl:padyl + lny,
-                padxl:padxl + lnx
-            ]
+            filtered_proj[p, :, :] = inarr[padyl:padyl+lny, padxl:padxl+lnx]
 
         # The Cartesian axes in our array are ordered like this: [z,y,x]
         # However, the rotation matrix requires [x,y,z]. Therefore, we
@@ -992,12 +947,4 @@ def backpropagate_3d_tilted(uSin, angles, res, nm, lD=0,
         if count is not None:
             count.value += 1
 
-    pool4loop.terminate()
-    pool4loop.join()
-
-    del _shared_array, inarr, odtbrain._shared_array
-    del shared_array_base
-
-    gc.collect()
-
     return outarr
